//******************************************************************************************
//* Asteroids, by Paul Nylander, bugman123.com, 10/12/04
//******************************************************************************************

// In order to compile this, you might need these OpenGL files: http://www.ics.uci.edu/~frost/past/ics186a/EssentialFiles.zip
// Simply place the files in these folders: http://www.ics.uci.edu/~frost/past/ics186a/wheretheygo.html

#include <GL/glut.h>
#include <stdio.h> // sprintf
#include <stdlib.h> // malloc
#include <math.h> // atan2, sqrt
#include <string.h> // strlen
#include <windows.h> // MessageBox

#define pi 3.14159 // 3.14159265358979323846

int xscreen=800, yscreen=600, window; // 1152x864, 800x600, 720x480, 320x240
int ibullet=0, iexpl=0, nbomb, pause=1, level, nroid, score;
double seed=pi;

struct Body {double x, y, vx, vy; int energy, dir;}; // class
typedef struct Body Body;
Body ship, enemy, planet, *bullets, *asteroids, *explosions;

struct Image {int xmax,ymax, transparent; char *pixels;};
typedef struct Image Image;
Image *planetImage, *shipImages[16], *enemyImages[16], *asteroidImages[3], *explosionImages[7];

struct Texture {int xmax,ymax; GLuint ID;};
typedef struct Texture Texture;
Texture background, planetTexture;
//Texture *shipTextures, *enemyTextures, *asteroidTextures, *explosionTextures;

double random(void) {int i; seed=1000*pi*seed; i=seed; seed-=i; return seed;}
double sqr(double x) {return x*x;};

static unsigned int getshort(file) FILE *file; {
 return (unsigned int)getc(file)+(((unsigned int)getc(file))<<8);
}
static unsigned int getint(file) FILE *file; {return
 (unsigned int)getc(file)+
 (((unsigned int)getc(file))<<8)+
 (((unsigned int)getc(file))<<16)+
 (((unsigned int)getc(file))<<24);
}
Image *readbmp(int transparent, char *filename) {
 int i,imax, components=transparent?4:3; // unsigned long
 char t;
 FILE *file;
 Image *image=(Image *)malloc(sizeof(Image));
 image->transparent=transparent;
 if(image==NULL) {exit(0);} // insufficient memory
 if((file=fopen(filename,"rb"))==NULL) {exit(0);} // file not found
 fseek(file,18,SEEK_CUR);
 image->xmax=getint(file);
 image->ymax=getint(file);
 imax=image->xmax*image->ymax;
 if(getshort(file)!=1) {exit(0);} // must have 1 plane
 if(getshort(file)!=24) {exit(0);} // must be 24 bit
 fseek(file,24,SEEK_CUR);
 image->pixels=(char *)malloc(components*imax);
 if(image->pixels==NULL) {exit(0);} // insufficient memory for image pixels
 if(fread(image->pixels,3*imax,1,file)!=1) {exit(0);} // Error reading image pixels
 for(i=0; i<3*imax; i+=3) { // swap colors, bgr -> rgb
  t=image->pixels[i];
  image->pixels[i]=image->pixels[i+2];
  image->pixels[i+2]=t;
 }
 if(transparent) {for(i=imax-1; i>=0; i--) { // insert alpha in pixels // http://zoo.cs.yale.edu/classes/cs490/02-03b/robert.rodi/source/texture.cpp
  image->pixels[4*i]=image->pixels[3*i];
  image->pixels[4*i+1]=image->pixels[3*i+1];
  image->pixels[4*i+2]=image->pixels[3*i+2];
  image->pixels[4*i+3]=(image->pixels[3*i]==-64 && image->pixels[3*i+1]==-64 && image->pixels[3*i+2]==-64)?0:255; // ''
 }}
 return image;
}

void drawImage(double x, double y, Image *image) {
 glRasterPos2i((int)(x-0.5*image->xmax),(int)(y-0.5*image->ymax)); // why doesn't glRasterPos2f work?
 glDrawPixels(image->xmax,image->ymax,image->transparent?GL_RGBA:GL_RGB,GL_UNSIGNED_BYTE,image->pixels); // glBitmap: monotone?
}

void makeTexture(int background, Image *image, Texture *texture) { // convert image to texture
 int format=(image->transparent)?GL_RGBA:GL_RGB; // components // GL_ALPHA
 texture->xmax=image->xmax; texture->ymax=image->ymax;
 glGenTextures(1,&texture->ID);
//glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST); // GL_LINEAR
 glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST); // GL_LINEAR // needed for laptop
 glBindTexture(GL_TEXTURE_2D,texture->ID); // redundant with glTexImage2D?
//if(background) { // how do I make a tiling?
// glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
// glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
//}
 glTexImage2D(GL_TEXTURE_2D,0,format,image->xmax,image->ymax,0,format,GL_UNSIGNED_BYTE,image->pixels);
 free(image->pixels);
 free(image);
}

void drawTexture(int bg, double x, double y, Texture texture) {
 double x1=bg?-1:((2*x-texture.xmax)/xscreen-1), y1=bg?-1:((2*y-texture.ymax)/yscreen-1);
 double x2=bg? 1:((2*x+texture.xmax)/xscreen-1), y2=bg? 1:((2*y+texture.ymax)/yscreen-1);
 glBindTexture(GL_TEXTURE_2D,texture.ID);
 glBegin(GL_QUADS);
 glTexCoord2f(0,0); glVertex2f(x1,y1);
 glTexCoord2f(1,0); glVertex2f(x2,y1);
 glTexCoord2f(1,1); glVertex2f(x2,y2);
 glTexCoord2f(0,1); glVertex2f(x1,y2);
 glEnd();
//glInterleavedArrays(GL_V2F,0,{{x1,y1},{x2,y1},{x2,y2},{x1,y2}});
//glDrawArrays(GL_QUADS,0,4);
}

void text(int x, int y, char *txt) {
 int n=(int)strlen(txt), i;
 glRasterPos2f(x,y);
 for(i=0; i<n; i++) {glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, txt[i]);}
}

double gettheta(int dir) {return 2*pi*(4-dir)/16.0;};
double torus(double x, double xmax) { // 2 Torus
 double x2=x<0?xmax:(x>xmax?0:x); // why do I need this?
 return x2;
};
int collision(double x1, double y1, double x2, double y2, double r) {
 return (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)<r*r;
};
void explosion(double x, double y) {
 iexpl=(iexpl+1)%16;
 explosions[iexpl].energy=7;
 explosions[iexpl].x=x;
 explosions[iexpl].y=y;
}
void fire(double theta) {
 ibullet=(ibullet+1)%32;
 bullets[ibullet].energy=500;
 bullets[ibullet].x=ship.x;
 bullets[ibullet].y=ship.y; // error: somehow explosions[2].energy is getting changed here
 bullets[ibullet].vx=ship.vx+1.0*cos(theta);
 bullets[ibullet].vy=ship.vy+1.0*sin(theta);
//bullets[ibullet].dir=ship.dir;
}
void newGame(int reset) {
 int i, start;
 if(reset) {level=1; score=0; nbomb=3;}
 nroid=(level>7)?256:(2<<level); // nroid doubles each time
 planet.x=0.5*xscreen; planet.y=0.5*yscreen;
 ship.x=planet.x+128; ship.y=planet.y; ship.vx=0; ship.vy=sqrt(10.0/128); ship.energy=1; ship.dir=0;
 start=1;
 enemy.vx=0; enemy.vy=0; enemy.energy=level; enemy.dir=0;
 while(start || collision(enemy.x,enemy.y,ship.x,ship.y,16+128+16)) {
  enemy.x=xscreen*random(); enemy.y=yscreen*random(); start=0;
 }
 for(i=0; i<nroid; i++) {start=1;
  while(start || collision(asteroids[i].x,asteroids[i].y,ship.x,ship.y,16+128+32)) {
   asteroids[i].x=xscreen*random(); asteroids[i].y=yscreen*random(); start=0;
  };
  asteroids[i].vy=2*random()-1; asteroids[i].vx=2*random()-1;
  asteroids[i].energy=3;
 }
 for(i=0; i<32; i++) {bullets[i].energy=0;}
 for(i=0; i<16; i++) {explosions[i].energy=0;}
}

// Calculate Animation
void animate(void) {int i,j; double theta,dv;
 if (!pause) {

  // Accelleration
  if(ship.energy>0) {
   theta=atan2(planet.y-ship.y,planet.x-ship.x);
   dv=10.0/(sqr(planet.y-ship.y)+sqr(planet.x-ship.x)+1); // sqrt(planet.y-ship.y,planet.x-ship.x)+1;
   ship.vx+=dv*cos(theta); ship.vy+=dv*sin(theta); // gravity toward planet
   ship.x=torus(ship.x+ship.vx,xscreen);
   ship.y=torus(ship.y+ship.vy,yscreen);
  }
  if(enemy.energy>0) {
   theta=atan2(ship.y-enemy.y,ship.x-enemy.x);
   enemy.dir=(16+4-(int)(8*(theta/pi)))%16;
   enemy.vx+=0.005*cos(theta); enemy.vy+=0.005*sin(theta);
   theta=atan2(planet.y-enemy.y,planet.x-enemy.x);
   dv=10.0/(sqr(planet.y-enemy.y)+sqr(planet.x-enemy.x)+1);
   enemy.vx+=dv*cos(theta); enemy.vy+=dv*sin(theta); // gravity toward planet
   enemy.x=torus(enemy.x+enemy.vx,xscreen);
   enemy.y=torus(enemy.y+enemy.vy,yscreen);
  }

  // Bullets
  for(i=0; i<32; i++) {if(bullets[i].energy>0) {
   bullets[i].x=torus(bullets[i].x+bullets[i].vx,xscreen);
   bullets[i].y=torus(bullets[i].y+bullets[i].vy,yscreen); // error: somehow explosions[i].energy is getting changed here
   bullets[i].energy--;
   if(bullets[i].energy>0 && enemy.energy>0 && collision(bullets[i].x,bullets[i].y,enemy.x,enemy.y,16)) {
    enemy.energy--; bullets[i].energy=0;
    if(enemy.energy==0) {nbomb++; score+=1000;} else score+=100;
    explosion(bullets[i].x,bullets[i].y);
   };
   for(j=0; j<nroid; j++) {
    if(bullets[i].energy>0 && asteroids[j].energy>0 && collision(bullets[i].x,bullets[i].y,asteroids[j].x,asteroids[j].y,4<<asteroids[j].energy)) {
     asteroids[j].energy--; bullets[i].energy=0;
     if(asteroids[j].energy==0) {nbomb++; score+=1000;} else score+=100;
     explosion(bullets[i].x,bullets[i].y);
    }
   }
  }}

  // Collisions
  if(ship.energy>0 && enemy.energy>0) {
   if(collision(enemy.x,enemy.y,ship.x,ship.y,16+16) && ship.energy>0) {
    ship.energy--; enemy.energy--;
    explosion(ship.x,ship.y);
   };
  }
  for(i=0; i<nroid; i++) {if(asteroids[i].energy>0) {
   theta=atan2(planet.y-asteroids[i].y,planet.x-asteroids[i].x);
   dv=10.0/(sqr(planet.y-asteroids[i].y)+sqr(planet.x-asteroids[i].x)+1);
   asteroids[i].vx+=dv*cos(theta); asteroids[i].vy+=dv*sin(theta);
   asteroids[i].x=torus(asteroids[i].x+asteroids[i].vx,xscreen);
   asteroids[i].y=torus(asteroids[i].y+asteroids[i].vy,yscreen);
   if(ship.energy>0 && collision(asteroids[i].x,asteroids[i].y,ship.x,ship.y,(4<<asteroids[i].energy)+16)) {
    ship.energy--; asteroids[i].energy--;
    explosion(ship.x,ship.y);
   };
  }}
 }
 glutPostRedisplay();
}

void thrust(int key, int xmouse, int ymouse) {
 double theta=gettheta(ship.dir);
 switch(key) {
  case GLUT_KEY_UP: pause=0; ship.vx+=0.05*cos(theta); ship.vy+=0.05*sin(theta); break;
  case GLUT_KEY_LEFT: pause=0; ship.dir=(ship.dir+15)%16; break;
  case GLUT_KEY_RIGHT: pause=0; ship.dir=(ship.dir+1)%16; break;
 }
 animate();
}

void keyboard(unsigned char key, int xmouse, int ymouse) {double theta; switch (key) {
 case 'n': case 'N': newGame(1); break;
 case 'h': case 'H': case '\r': pause=!pause; break;
 case 'q': case 'Q': case 27: glutDestroyWindow(window); exit(0); break;
 case ' ': pause=0; if(ship.energy>0) {fire(gettheta(ship.dir)); animate();} break;
 case 'b': case 'B': pause=0;
  if((ship.energy>0) && nbomb>0) {nbomb--;
   for(theta=0; theta<2*pi; theta+=pi/8) {fire(theta);}
   animate();
  }
 break;
}}

//void resize(int x, int y) {xscreen=x; yscreen=y;} // doesn't help

void display(void) {
 int i, win=1;
 char txt[100];
 glClear(GL_COLOR_BUFFER_BIT);
 glLoadIdentity();

 glEnable(GL_TEXTURE_2D);
 drawTexture(1,0,0,background); // Draw Space Background
//drawTexture(0,planet.x,planet.y,planetTexture); // Draw Planet, why doesn't this work?
 glDisable(GL_TEXTURE_2D);

 gluOrtho2D(0,xscreen,0,yscreen);

 drawImage(planet.x,planet.y,planetImage); // Draw Planet

 // Draw Ship
 if(ship.energy>0) {
//drawTexture(0,ship.x,ship.y,shipTextures[ship.dir]);
  drawImage(ship.x,ship.y,shipImages[ship.dir]);
 }

 // Draw Enemy Rocket
 if(enemy.energy>0) {
//drawTexture(0,enemy.x,enemy.y,enemyImages[enemy.dir]);
  drawImage(enemy.x,enemy.y,enemyImages[enemy.dir]);
 }

 // Draw Asteroids
 for(i=0; i<nroid; i++) {if(asteroids[i].energy>0) {win=0;
//drawTexture(0,asteroids[i].x,asteroids[i].y,asteroidTextures[asteroids[i].energy-1]);
  drawImage(asteroids[i].x,asteroids[i].y,asteroidImages[asteroids[i].energy-1]);
 }}

 // Draw Explosions
 for(i=0; i<16; i++)
  {if(explosions[i].energy>0 && explosions[i].energy<=7) { // should never exceed 7
// drawTexture(0,explosions[i].x,explosions[i].y,explosionTextures[explosions[i].energy-1]);
   drawImage(explosions[i].x,explosions[i].y,explosionImages[explosions[i].energy-1]);
   explosions[i].energy--;
  }
 }

 // Draw Bullets
//glColor3f(1,1,1);
 for(i=0; i<32; i++) {if(bullets[i].energy>0) {
  glPointSize((GLfloat)(0.01*(bullets[i].energy)));
  glBegin(GL_POINTS);
  glVertex2i((int)bullets[i].x,(int)bullets[i].y);
  glEnd();
 }}

 // Score
 glPointSize(5);
 glBegin(GL_POINTS);
 for(i=0; i<nbomb; i++) {glVertex2i(10*(i+1),25);}
 glEnd();
 sprintf(txt,"Level: %i, Score: %i",level,score); text(10,10,txt);
 if(pause) {
  text(xscreen/2-110,yscreen/2+7,"Use the arrow keys to move your ship.");
  text(xscreen/2-110,yscreen/2-5,"Press the space bar to fire, 'B' for Bomb");
  text(xscreen/2-120,yscreen/2-17,"'N' for New Game, 'H' for Help, 'Q' to Quit."); // 'P' for Pause
 }
 else if(win) {level++; newGame(0);} // text(0.5*xscreen-20,0.5*yscreen-10,"You Win!");
 else if(ship.energy<=0) {text(xscreen/2-20,yscreen/2-15,"Game Over");}

 glutSwapBuffers();
}

void main(int argc, char** argv) {
 int i, fullscreen=(MessageBox(NULL,"Fullscreen mode?","Asteroids",MB_YESNO)!=IDNO);
 char filename[100]; // 12?
 bullets=(Body *)malloc(32);
 asteroids=(Body *)malloc(256); // max possible asteroids (highly unlikely)
 explosions=(Body *)malloc(16);
 newGame(1); // xscreen, yscreen might not always be right

 // Graphics
 glutInit(&argc,argv);
 glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
 if(fullscreen) { // glutGameModeString("360x240:32");
  glutEnterGameMode();
//xscreen=glutGameModeGet(GLUT_GAME_MODE_WIDTH); // 1152 // crashes!
//yscreen=glutGameModeGet(GLUT_GAME_MODE_HEIGHT); // 864
//newGame(1);
 }
 else {
  glutInitWindowSize(xscreen,yscreen);
  glutInitWindowPosition(0,0);
  window=glutCreateWindow("Asteroids");
 }
 glutSetCursor(GLUT_CURSOR_NONE); // hide cursor
 glClearColor(0,0,0,1); // black background
 glutDisplayFunc(display);
//glutReshapeFunc(resize);
 glutIdleFunc(animate);
 glutSpecialFunc(thrust);
 glutKeyboardFunc(keyboard);
 glAlphaFunc(GL_NOTEQUAL,0); // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
 glEnable(GL_ALPHA_TEST);
//glEnable(GL_SCISSOR_TEST); // messes everything up, see what laptop does, http://www.parallab.uib.no/SGI_bookshelves/SGI_Developer/books/OpenGL_PG/sgi_html/ch11.html#id97883
//glScissor(0,0,xscreen,yscreen);

 // Read Images and Textures
 makeTexture(1,readbmp(0,"Background.bmp"),&background);
 planetImage=readbmp(1,"Planet.bmp");
 makeTexture(0,readbmp(1,"Planet.bmp"),&planetTexture);
//shipTextures=(Texture *)malloc(16);
 for(i=0; i<16; i++) {
  sprintf(filename,"Ships/%i.bmp",i);
  shipImages[i]=readbmp(1,filename);
//makeTexture(0,readbmp(1,filename),&shipTextures[i]);
 }
//enemyTextures=(Texture *)malloc(16);
 for(i=0; i<16; i++) {
  sprintf(filename,"Enemy/%i.bmp",i);
  enemyImages[i]=readbmp(1,filename);
//makeTexture(0,readbmp(1,filename),&enemyTextures[i]);
 }
//asteroidTextures=(Texture *)malloc(3);
 for(i=0; i<3; i++) {
  sprintf(filename,"Asteroids/%i.bmp",i+1);
  asteroidImages[i]=readbmp(1,filename);
//makeTexture(0,readbmp(1,filename),&asteroidTextures[i]);
 }
//explosionTextures=(Texture *)malloc(7);
 for(i=0; i<7; i++) {
  sprintf(filename,"Explosion/%i.bmp",i+1);
  explosionImages[i]=readbmp(1,filename);
//makeTexture(0,readbmp(1,filename),&explosionTextures[i]);
 }

 glutMainLoop();
}
